import math
import random
from typing import List, Tuple

from shor.gates import X, CNOT, CSWAP, H, Rx, QFT
from shor.layers import Qubits
from shor.operations import Measure
from shor.quantum import Circuit


def factor(N: int) -> Tuple[int, int]:
    """ WIP not currently functioning.

    The classical part seems to be correct, but the quantum period finding function is incorrect
    at the moment.
    """
    while True:
        a = random.randint(2, N - 1)
        common_divisor = gcd(a, N)
        if common_divisor > 1:
            # Lucky guess
            return common_divisor, N / common_divisor

        # r = shor.quantum.shor.period(a, N)
        r = find_period(a, N)

        a_pow_r_o_2 = a ^ (r / 2)

        # if r is odd or
        if r & 1 == 1 or a_pow_r_o_2 == (-1 % N):
            continue

        factor = max(gcd(a_pow_r_o_2 + 1, N), gcd(a_pow_r_o_2 - 1, N))
        return factor, N / factor


def find_period(a, N):
    """ WIP: Quantum subroutine for shor's algorithm.
    Finds the period of a function of the form:
    f(x) = a^x % N

    This uses the quantum fourier transform.
    """

    circuit = Circuit()
    circuit.add(Qubits(5))
    circuit.add(QFT(0, 1, 2, 3))
    circuit.add(quantum_amod_15(a))
    circuit.add(QFT(3, 2, 1, 0))  # Inverse Quantum Fourier transform

    from shor.backends import QuantumSimulator, QSession

    sess = QSession(backend=QuantumSimulator())
    result = sess.run(circuit, num_shots=1024)
    return result


def quantum_amod_15(a: int) -> Circuit:
    qc = Circuit()

    if a == 2:
        qc.add(CSWAP(4, 3, 2))
        qc.add(CSWAP(4, 2, 1))
        qc.add(CSWAP(4, 1, 0))
    if a == 4 or a == 11 or a == 14:
        qc.add(CSWAP(4, 2, 0))
        qc.add(CSWAP(4, 3, 1))
        qc.add(CNOT(4, 3))
        qc.add(CNOT(4, 2))
        qc.add(CNOT(4, 1))
        qc.add(CNOT(4, 0))
    if a == 7:
        qc.add(CSWAP(4, 1, 0))
        qc.add(CSWAP(4, 2, 1))
        qc.add(CSWAP(4, 3, 2))
        qc.add(CNOT(4, 3))
        qc.add(CNOT(4, 2))
        qc.add(CNOT(4, 1))
        qc.add(CNOT(4, 0))
    if a == 8:
        qc.add(CSWAP(4, 1, 0))
        qc.add(CSWAP(4, 2, 1))
        qc.add(CSWAP(4, 3, 2))
    if a == 13:
        qc.add(CSWAP(4, 3, 2))
        qc.add(CSWAP(4, 2, 1))
        qc.add(CSWAP(4, 1, 0))
        qc.add(CNOT(4, 3))
        qc.add(CNOT(4, 2))
        qc.add(CNOT(4, 1))
        qc.add(CNOT(4, 0))

    return qc


def qft(qubits: List[int]) -> Circuit:
    qc = Circuit()
    for i in range(len(qubits)):
        for k in range(i):
            qc.add(Rx(qubits[i], qubits[k], angle=math.pi/float(2**(i-k))))
        qc.add(H(qubits[i]))
    return qc


def gcd(a: int, b: int) -> int:
    """Finds greatest common divisor between positive integers a and b
    Efficient, using Euclidean algorithm.

    Example:
    >>> greatest_common_divisor(15, 25)
    5
    """
    assert a > 0 and b > 0, 'Inputs should be positive integers'

    to_divide, remainder = max(a, b), min(a, b)  # initial divisor
    divisor = 1

    while remainder != 0:
        divisor = remainder
        remainder = to_divide
        while remainder >= divisor:
            remainder -= divisor
        to_divide = divisor

    return divisor



from shor.quantum import Circuit

import os
import pydot


def plot_circuit(circuit: Circuit, to_file='model.png'):
    """ Converts a Quantum Circuit into a diagram, and saves to a file

    Args:
        circuit: A quantum :func:`Circuit() <shor.quantum.Circuit>` object
        to_file: File name of the plot image.

    Returns:
        A Jupyter notebook Image object if Jupyter is installed.
        This enables in-line display of the circuit diagram in notebooks.
    """
    pass


def circuit_to_dot(circuit: Circuit):
    """ Converts a Quantum Circuit into Dot format
    See: pydot

    Args:
        circuit: A shor.quantum.Circuit object
    """
    pass


from __future__ import annotations
from typing import List, Union


from shor.layers import _BaseLayer, Qubits
from shor.operations import Measure


class Circuit(object):
    def __init__(self):
        self.layers: List[_BaseLayer] = []

    def add(self, layer_or_circuit: Union[_BaseLayer, Circuit]):

        if isinstance(layer_or_circuit, _BaseLayer):
            self.layers.append(layer_or_circuit)
        elif isinstance(layer_or_circuit, Circuit):
            self.layers.extend(layer_or_circuit.layers)
        else:
            raise TypeError("Circuit class cannot add the type: {}".format(type(layer_or_circuit)))

        return self

    def initial_state(self):
        initial_qubits = []
        for qbit_layer in filter(lambda l: type(l) == Qubits, self.layers):
            initial_qubits.extend([qbit_layer.state] * qbit_layer.num)

        return initial_qubits

    def to_gates(self):
        gates = []
        for l in self.layers:
            gates.extend(l.to_gates())
        return gates

    def measure_bits(self):
        measure_bits = []
        for m in filter(lambda l: type(l) == Measure, self.layers):
            measure_bits.extend(m.bits)
        return measure_bits



from collections import deque
from functools import lru_cache
from typing import List, Deque, NamedTuple, Tuple, Dict

import numpy as np

from shor.gates import _Gate
from shor.quantum import Circuit


class _GateTuple(NamedTuple):
    qubits: List[int]
    matrix: np.ndarray


class _QuantumBackend(object):
    def run(self, initial_state, gates, measure_bits):
        pass


class QuantumSimulator(_QuantumBackend):

    @staticmethod
    @lru_cache(maxsize=10)
    def combine_gates(gates: Tuple[_Gate]) -> _GateTuple:
        to_combine: Deque[_GateTuple] = deque(_GateTuple(g.qubits, g.to_matrix()) for g in gates)
        combined: Deque[_GateTuple] = deque()

        while len(to_combine) > 1:
            while len(to_combine) > 1:
                gate1 = to_combine.popleft()
                gate2 = to_combine.popleft()

                can_combine_no_tensor = set(gate1.qubits) == set(gate2.qubits)
                can_combine_with_tensor = not any(q in gate2.qubits for q in gate1.qubits)

                if can_combine_no_tensor:
                    combined.append(_GateTuple(
                        gate1.qubits,
                        np.dot(
                            gate1.matrix,
                            change_qubit_order(gate2.matrix, gate2.qubits, gate1.qubits)
                        )
                    ))
                elif can_combine_with_tensor:
                    combined.append(_GateTuple(
                        gate1.qubits + gate2.qubits,
                        np.kron(
                            gate1.matrix,
                            gate2.matrix
                        )
                    ))
                else:
                    common_qubits = [q for q in gate1.qubits if q in gate2.qubits]
                    q1_only_qubits = [q for q in gate1.qubits if q not in common_qubits]
                    q2_only_qubits = [q for q in gate2.qubits if q not in common_qubits]
                    all_qubits = q1_only_qubits + common_qubits + q2_only_qubits

                    # Gate 1
                    # Rearrange gates so common qubits are on right side of gate 1
                    new_gate1_qubit_order = q1_only_qubits + common_qubits
                    new_gate1 = change_qubit_order(gate1.matrix, gate1.qubits, new_gate1_qubit_order)

                    # Tensor product with I to fill remaining qubits
                    num_qubits_to_add = len(all_qubits) - len(gate1.qubits)
                    if num_qubits_to_add > 0:
                        new_gate1 = np.kron(new_gate1, np.eye(np.power(2, num_qubits_to_add)))

                    # Gate 2
                    # Rearrange gates so common qubits are on left side of gate 2
                    new_gate2_qubit_order = common_qubits + q2_only_qubits
                    new_gate2 = change_qubit_order(gate2.matrix, gate2.qubits, new_gate2_qubit_order)

                    # Tensor product with I to fill remaining qubits
                    num_qubits_to_add = len(all_qubits) - len(gate2.qubits)
                    if num_qubits_to_add > 0:
                        new_gate2 = np.kron(np.eye(np.power(2, num_qubits_to_add)), new_gate2)

                    combined.append(_GateTuple(
                        all_qubits,
                        np.dot(
                            new_gate1,
                            new_gate2
                        )
                    ))

            to_combine = combined

        return to_combine.pop()

    def run(self, initial_state: np.ndarray, gates: List[_Gate], measure_bits: List[int]):
        combined = self.combine_gates(tuple(gates))
        new_qubit_order = combined.qubits
        state_vector = get_entangled_initial_state(initial_state, new_qubit_order)

        probabilities = np.square(np.dot(state_vector, combined.matrix))

        return np.random.choice(state_vector.shape[0], p=probabilities)


class QSession:
    def __init__(self, **kwargs):
        self.backend = kwargs.get('backend', QuantumSimulator())

    def run(self, circuit: Circuit, **kwargs):
        num_shots = kwargs.get('num_shots', 1024)

        initial_state = circuit.initial_state()
        gates = circuit.to_gates()
        measure_bits = circuit.measure_bits()

        counts = {}

        for i in range(num_shots):
            state = self.backend.run(initial_state, gates, measure_bits)
            counts[state] = counts.get(state, 0) + 1

        return QResult(counts)


class QCounts:
    def __init__(self, counts: Dict[int, int]):
        self.counts: Dict[int, int] = counts

    def get(self, key, default=0):
        if isinstance(key, str):
            idx = int_from_bit_string(key)
        else:
            idx = key

        return self.counts.get(idx, default)


class QResult:
    def __init__(self, counts: Dict[int, int]):
        self.counts: QCounts = QCounts(counts)


def get_entangled_initial_state(initial_state, new_qubit_order):
    states_to_entangle: Deque[np.ndarray] = deque()

    states = list(map(lambda s: np.asarray([1 if s == 0 else 0, 1 if s == 1 else 0]), initial_state))

    for q in new_qubit_order:
        states_to_entangle.append(states[q])

    while len(states_to_entangle) > 1:
        entangled_states: Deque[np.ndarray] = deque()
        while len(states_to_entangle) > 1:
            entangled_states.append(np.kron(states_to_entangle.popleft(), states_to_entangle.popleft()))
        if len(states_to_entangle) == 1:
            entangled_states.append(states_to_entangle.popleft())
        states_to_entangle = entangled_states

    return states_to_entangle.pop()


def change_qubit_order(matrix: np.ndarray, old_order, new_order) -> np.ndarray:
    reorder_qubits = [new_order.index(q2) for q2 in old_order]
    initial_row_order = [i for i in range(np.power(2, len(old_order)))]

    reorder_rows = list(map(lambda i: rearrange_bits(i, reorder_qubits), initial_row_order))

    return matrix[reorder_rows][:, reorder_rows]


def rearrange_bits(num: int, new_bit_order: List[int]):
    sig_bits = len(new_bit_order)
    bit_string = int_to_bits(num, sig_bits)[2:]
    result = int('0b' + ''.join(bit_string[-b] for b in new_bit_order), 2)
    return result


def int_to_bits(num: int, sig_bits: int) -> str:
    return format(num, '#0%sb' % np.power(2, sig_bits))


def int_from_bit_string(bits: str) -> int:
    return int(bits, 2) if bits.startswith('0b') else int('0b' + bits, 2)




    # Accounting for random noise, results won't be exact
    assert result.counts.get(bin(0)) > 450
    assert result.counts.get(bin(1)) > 450


def test_entanglement():
    circuit = Circuit()
    circuit.add(Qubits(2))
    circuit.add(Hadamard(0))
    circuit.add(CNOT(0, 1))
    circuit.add(Measure(0, 1))

    sess = QSession(backend=QuantumSimulator())
    result = sess.run(circuit, num_shots=1024)

    # Accounting for random noise, results won't be exact
    # For result index:
    #     - int('10',2)
    #     - 0b10
    #     - 2
    # are all equivalant ways to get the index for '|10>'
    assert result.counts.get(bin(1)) == 0
    assert result.counts.get(0b10) == 0
    assert result.counts.get('00') > 450
    assert result.counts.get(3) > 450


def test_unitary_symmetry_does_nothing():
    symmetric_circuit_1 = Circuit()
    symmetric_circuit_1.add(Qubits(2))
    symmetric_circuit_1.add(Hadamard(0))
    symmetric_circuit_1.add(Hadamard(0))
    symmetric_circuit_1.add(CNOT(0, 1))
    symmetric_circuit_1.add(CNOT(0, 1))
    symmetric_circuit_1.add(Measure(0, 1))

    symmetric_circuit_2 = Circuit()
    symmetric_circuit_2.add(Qubits(2, state=1))
    symmetric_circuit_2.add(Hadamard(0))
    symmetric_circuit_2.add(CNOT(0, 1))
    symmetric_circuit_2.add(CNOT(0, 1))
    symmetric_circuit_2.add(Hadamard(0))
    symmetric_circuit_2.add(Measure(0, 1))

    sess = QSession(backend=QuantumSimulator())
    result_1 = sess.run(symmetric_circuit_1, num_shots=1024)
    result_2 = sess.run(symmetric_circuit_2, num_shots=1024)

    assert result_1.counts.get(0) == 1024
    assert result_2.counts.get(3) == 1024



from shor.gates import H
from shor.layers import Qbits


def test_quantum_backend_init():
    from shor.backends import _QuantumBackend
    _QuantumBackend()


def test_quantum_simulator_init():
    from shor.backends import QuantumSimulator
    QuantumSimulator()


def test_qsession_init():
    from shor.backends import QSession
    QSession()


def test_qsession_run():
    from shor.backends import QSession
    sess = QSession()

    from shor.quantum import Circuit
    sess.run(Circuit().add(Qbits(1)).add(H()), num_shots=10)



from shor.algorithms.shor import factor
import random

# Seed for tests
random.seed(0)


def test_shor_factoring():
    N = 3 * 5

    a, b = factor(N)

    assert a == 3 or a == 5
    assert b == N / a

    N = 3 * 7

    a, b = factor(N)

    assert a == 3 or a == 7
    assert b == N / a



from setuptools import setup, find_packages

with open("README.md", "r") as readme_file:
    readme = readme_file.read()

requirements = []

setup(
    name="shor",
    version="0.0.2a1",
    author="shor.dev",
    author_email="shordotdev@gmail.com",
    description="Quantum Computing for Humans",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/shor-team/shor",
    packages=find_packages(),
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
    ],
)
